@@ -1,9 +1,9 @@
 /*
  *  Pcap.c
  *
- *  $Id: Pcap.c,v 1.6 1999/11/08 06:40:17 fukusima Exp $
+ *  $Id: Pcap.c,v 1.10 2000/08/13 05:56:31 fukusima Exp $
  *
- *  Copyright (C) 1998, 1999  Masaki Fukushima
+ *  Copyright (C) 1998-2000  Masaki Fukushima
  */
 
 #include "ruby_pcap.h"
@@ -12,6 +12,7 @@
 #include <sys/types.h>
 #include <unistd.h>
 
+#define DEFAULT_DATALINK	DLT_EN10MB
 #define DEFAULT_SNAPLEN	68
 #define DEFAULT_PROMISC	1
 #define DEFAULT_TO_MS	1000
@@ -20,10 +21,24 @@ static char pcap_errbuf[PCAP_ERRBUF_SIZE];
 VALUE mPcap, rbpcap_convert = Qnil;
 VALUE ePcapError;
 VALUE eTruncatedPacket;
+VALUE cFilter;
 static VALUE cCapture;
 static VALUE cPcapStat;
 static VALUE cDumper;
-static VALUE cFilter;
+
+struct filter_object {
+    char *expr;
+    struct bpf_program program;
+    int datalink;
+    int snaplen;
+    VALUE param;
+    VALUE optimize;
+    VALUE netmask;
+};
+
+#define GetFilter(obj, filter) \
+    Data_Get_Struct(obj, struct filter_object, filter)
+
 
 static VALUE
 pcap_s_lookupdev(self)
@@ -71,10 +86,6 @@ pcap_s_convert_set(self, val)
     return Qnil;
 }
 
-#define CheckClass(v, class) ((RTEST(rb_obj_is_kind_of(v, class))) ? 0 :\
-    rb_raise(rb_eTypeError, "wrong type %s (expected %s)",\
-        rb_class2name(CLASS_OF(v)), rb_class2name(class)))
-
 /*
  * Capture object
  */
@@ -331,7 +342,8 @@ capture_setfilter(argc, argv, self)
      VALUE self;
 {
     struct capture_object *cap;
-    VALUE filter, optimize;
+    VALUE vfilter, optimize;
+    char *filter;
     int opt;
     struct bpf_program program;
 
@@ -339,16 +351,23 @@ capture_setfilter(argc, argv, self)
     GetCapture(self, cap);
 
     /* scan arg */
-    if (rb_scan_args(argc, argv, "11", &filter, &optimize) == 1) {
+    if (rb_scan_args(argc, argv, "11", &vfilter, &optimize) == 1) {
 	optimize = Qtrue;
     }
 
     /* check arg */
-    Check_Type(filter, T_STRING);
+    if (IsKindOf(vfilter, cFilter)) {
+	struct filter_object *f;
+	GetFilter(vfilter, f);
+	filter = f->expr;
+    } else {
+	Check_Type(vfilter, T_STRING);
+	filter = RSTRING(vfilter)->ptr;
+    }
     opt = RTEST(optimize);
 
     /* operation */
-    if (pcap_compile(cap->pcap, &program, RSTRING(filter)->ptr,
+    if (pcap_compile(cap->pcap, &program, filter,
 		     opt, cap->netmask) < 0)
 	rb_raise(ePcapError, "setfilter: %s", pcap_geterr(cap->pcap));
     if (pcap_setfilter(cap->pcap, &program) < 0)
@@ -509,17 +528,16 @@ dumper_dump(self, v_pkt)
  * Filter object
  */
 
-struct filter_object {
-    char *expr;
-    struct bpf_program program;
-    int datalink;
-    int snaplen;
-};
-
-#define GetFilter(obj, filter) \
-    Data_Get_Struct(obj, struct filter_object, filter)
-
 /* called from GC */
+static void
+mark_filter(filter)
+     struct filter_object *filter;
+{
+    rb_gc_mark(filter->param);
+    rb_gc_mark(filter->optimize);
+    rb_gc_mark(filter->netmask);
+}
+
 static void
 free_filter(filter)
      struct filter_object *filter;
@@ -543,18 +561,29 @@ filter_new(argc, argv, class)
     struct capture_object *capture;
     pcap_t *pcap;
     char *expr;
-    int n, optimize;
+    int n, optimize, snaplen, linktype;
     bpf_u_int32 netmask;
 
-    n = rb_scan_args(argc, argv, "22", &v_expr, &v_capture,
+    n = rb_scan_args(argc, argv, "13", &v_expr, &v_capture,
 		     &v_optimize, &v_netmask);
     /* filter expression */
     Check_Type(v_expr, T_STRING);
     expr = STR2CSTR(v_expr);
     /* capture object */
-    CheckClass(v_capture, cCapture);
-    GetCapture(v_capture, capture);
-    pcap = capture->pcap;
+    if (IsKindOf(v_capture, cCapture)) {
+	CheckClass(v_capture, cCapture);
+	GetCapture(v_capture, capture);
+	pcap = capture->pcap;
+    } else if (NIL_P(v_capture)) {
+	/* assume most common case */
+	snaplen  = DEFAULT_SNAPLEN;
+	linktype = DEFAULT_DATALINK;
+	pcap = 0;
+    } else {
+	snaplen  = NUM2INT(rb_funcall(v_capture, rb_intern("[]"), 1, INT2FIX(0)));
+	linktype = NUM2INT(rb_funcall(v_capture, rb_intern("[]"), 1, INT2FIX(1)));
+	pcap = 0;
+    }
     /* optimize flag */
     optimize = 1;
     if (n >= 3) {
@@ -568,18 +597,32 @@ filter_new(argc, argv, class)
     }
 
     filter = (struct filter_object *)xmalloc(sizeof(struct filter_object));
-    if (pcap_compile(pcap, &filter->program, expr, optimize, netmask) < 0) {
-	rb_raise(ePcapError, "%s", pcap_geterr(pcap));
+    if (pcap) {
+	if (pcap_compile(pcap, &filter->program, expr, optimize, netmask) < 0)
+	    rb_raise(ePcapError, "%s", pcap_geterr(pcap));
+	filter->datalink = pcap_datalink(pcap);
+	filter->snaplen  = pcap_snapshot(pcap);
+    } else {
+#ifdef HAVE_PCAP_COMPILE_NOPCAP
+	if (pcap_compile_nopcap(snaplen, linktype, &filter->program, expr, optimize, netmask) < 0)
+	    /* libpcap-0.5 provides no error report for pcap_compile_nopcap */
+	    rb_raise(ePcapError, "pcap_compile_nopcap error");
+	filter->datalink = linktype;
+	filter->snaplen  = snaplen;
+#else
+	rb_raise(rb_eArgError, "pcap_compile_nopcap needs libpcap-0.5 or later");
+#endif
     }
-    self = Data_Wrap_Struct(class, 0, free_filter, filter);
-    filter->expr     = strdup(expr);
-    filter->datalink = pcap_datalink(pcap);
-    filter->snaplen  = pcap_snapshot(pcap);
+    self = Data_Wrap_Struct(class, mark_filter, free_filter, filter);
+    filter->expr	= strdup(expr);
+    filter->param	= v_capture;
+    filter->optimize	= optimize ? Qtrue : Qfalse;
+    filter->netmask	= INT2NUM(ntohl(netmask));
 
     return self;
 }
 
-static VALUE
+VALUE
 filter_match(self, v_pkt)
     VALUE self, v_pkt;
 {
@@ -592,16 +635,82 @@ filter_match(self, v_pkt)
     GetPacket(v_pkt, pkt);
     h = &pkt->hdr.pkthdr;
 
-    if (filter->datalink != pkt->hdr.dl_type
-	|| filter->snaplen < h->caplen) {
-	rb_raise(rb_eRuntimeError, "Incompatible Packet");
-    }
+    if (filter->datalink != pkt->hdr.dl_type)
+	rb_raise(rb_eRuntimeError, "Incompatible datalink type");
+    if (filter->snaplen < h->caplen)
+	rb_raise(rb_eRuntimeError, "Incompatible snaplen");
+
     if (bpf_filter(filter->program.bf_insns, pkt->data, h->len, h->caplen))
 	return Qtrue;
     else
 	return Qfalse;
 }
 
+static VALUE
+filter_source(self)
+    VALUE self;
+{
+    struct filter_object *filter;
+
+    GetFilter(self, filter);
+    return rb_str_new2(filter->expr);
+}
+
+static VALUE
+new_filter(expr, param, optimize, netmask)
+    char *expr;
+    VALUE param, optimize, netmask;
+{
+    return rb_funcall(cFilter,
+		      rb_intern("new"), 4,
+		      rb_str_new2(expr), param, optimize, netmask);
+}
+
+static VALUE
+filter_or(self, other)
+    VALUE self, other;
+{
+    struct filter_object *filter, *filter2;
+    char *expr;
+
+    CheckClass(other, cFilter);
+    GetFilter(self, filter);
+    GetFilter(other, filter2);
+
+    expr = ALLOCA_N(char, strlen(filter->expr) + strlen(filter2->expr) + 16); 
+    sprintf(expr, "( %s ) or ( %s )", filter->expr, filter2->expr);
+    return new_filter(expr, filter->param, filter->optimize, filter->netmask);
+}
+
+static VALUE
+filter_and(self, other)
+    VALUE self, other;
+{
+    struct filter_object *filter, *filter2;
+    char *expr;
+
+    CheckClass(other, cFilter);
+    GetFilter(self, filter);
+    GetFilter(other, filter2);
+
+    expr = ALLOCA_N(char, strlen(filter->expr) + strlen(filter2->expr) + 16); 
+    sprintf(expr, "( %s ) and ( %s )", filter->expr, filter2->expr);
+    return new_filter(expr, filter->param, filter->optimize, filter->netmask);
+}
+
+static VALUE
+filter_not(self)
+    VALUE self;
+{
+    struct filter_object *filter;
+    char *expr;
+
+    GetFilter(self, filter);
+    expr = ALLOCA_N(char, strlen(filter->expr) + 16); 
+    sprintf(expr, "not ( %s )", filter->expr);
+    return new_filter(expr, filter->param, filter->optimize, filter->netmask);
+}
+
 /*
  * Class definition
  */
@@ -664,6 +773,11 @@ Init_pcap(void)
     rb_define_singleton_method(cFilter, "compile", filter_new, -1);
     rb_define_method(cFilter, "=~", filter_match, 1);
     rb_define_method(cFilter, "===", filter_match, 1);
+    rb_define_method(cFilter, "source", filter_source, 0);
+    rb_define_method(cFilter, "|", filter_or, 1);
+    rb_define_method(cFilter, "&", filter_and, 1);
+    rb_define_method(cFilter, "~@", filter_not, 0);
+    /*rb_define_method(cFilter, "&", filter_and, 1);*/
 
     /* define class PcapStat */
     cPcapStat = rb_funcall(rb_cStruct, rb_intern("new"), 4,
